/*

*/

import renderer.scene.*;
import renderer.models.Disk;
import renderer.models.ModelShading;
import renderer.pipeline.*;
import renderer.framebuffer.*;
import renderer.gui.*;

import java.awt.Color;
import java.awt.event.*;

/**

*/
@SuppressWarnings("serial")
public class AntiAliasDemo extends AnimationFrame
{
   private Scene scene;
   private double rotation = 0.0;
   private double delta = 1.0;
   private Color backGroundColor = Color.black;

   /**
      This constructor instantiates the Scene object
      and initializes it with appropriate geometry.
   */
   public AntiAliasDemo(String title, int fbWidth, int fbHeight, int fps)
   {
      super(title, fbWidth, fbHeight, fps);

      Model disk = new Disk(1.0, 3, 8);
      ModelShading.setColor(disk, Color.yellow);

      // Create the Scene object that we shall render
      scene = new Scene();

      scene.camera.perspective = false;

      // Add the model to the Scene.
      scene.positionList.add(new Position(disk));
   }


   // Implement the ActionListener interface.
   @Override public void actionPerformed(ActionEvent e)
   {
      //System.out.println( e );

      Position disk = scene.positionList.get(0);
      // Set the model-to-view transformation matrix.
      disk.matrix2Identity();
      // Push the model away from where the camera is.
      disk.matrix.mult( Matrix.translate(0, 0, -8) );

      disk.matrix.mult( Matrix.rotateZ(rotation) );

      rotation += delta;

      // Render again.
      FrameBuffer fb = this.fbp.getFrameBuffer();
      fb.clearFB(backGroundColor);
      Pipeline.render(scene, fb.vp);
      fbp.update();
      repaint();
   }


   // Implement part of the KeyListener interface.
   @Override public void keyTyped(KeyEvent e)
   {
      //System.out.println( e );

      char c = e.getKeyChar();
      if ('h' == c)
      {
         print_help_message();
         return;
      }
      else if ('a' == c)
      {
         RasterizeAntialias.doAntialiasing = ! RasterizeAntialias.doAntialiasing;
         System.out.print("Anti-aliasing is turned ");
         System.out.println(RasterizeAntialias.doAntialiasing ? "On" : "Off");
      }
      else if ('g' == c)
      {
         RasterizeAntialias.doGamma = ! RasterizeAntialias.doGamma;
         System.out.print("Gamma correction is turned ");
         System.out.println(RasterizeAntialias.doGamma ? "On" : "Off");
      }
      else if ('b' == c)
      {
         // Change the FrameBuffer's background color.
         Model m = new Model();
         ModelShading.setRandomColor(m);
         backGroundColor = m.colorList.get(0);
      }
      else if ('c' == c)
      {
         // Change the solid random color of the disk.
         ModelShading.setRandomColor(scene.positionList.get(0).model);
      }
      else if ('C' == c)
      {
         // Change each color in the disk to a random color.
         ModelShading.setRandomColors(scene.positionList.get(0).model);
      }
      else if ('e' == c && e.isAltDown())
      {
         // Change the random color of each vertex of the disk.
         ModelShading.setRandomVertexColors(scene.positionList.get(0).model);
      }
      else if ('e' == c)
      {
         // Change the solid random color of each edge of the disk.
         ModelShading.setRandomLineSegmentColors(scene.positionList.get(0).model);
      }
      else if ('E' == c)
      {
         // Change the random color of each end of each edge of the disk.
         ModelShading.setRainbowLineSegmentColors(scene.positionList.get(0).model);
      }
      else if ('r' == c)
      {
         delta += 1.0;
         System.out.println("delta = " + delta);
      }
      else if ('R' == c)
      {
         delta -= 1.0;
         System.out.println("delta = " + delta);
      }
      else if ('f' == c)
      {
         int fps = getFPS();
         fps -= 1;
         if (0 == fps) fps = 1;
         setFPS(fps);
         System.out.println("fps = " + fps);
      }
      else if ('F' == c)
      {
         int fps = getFPS();
         fps += 1;
         setFPS(fps);
         System.out.println("fps = " + fps);
      }
   }


   // Implement part of the ComponentListener interface.
   @Override public void componentResized(ComponentEvent e)
   {
      //System.out.println( e );

      // Get the new size of the FrameBufferPanel.
      int w = this.fbp.getWidth();
      int h = this.fbp.getHeight();

      // Create a new FrameBuffer that fits the new window size.
      FrameBuffer fb = new FrameBuffer(w, h);
      this.fbp.setFrameBuffer(fb);
      fb.clearFB(backGroundColor);
      Pipeline.render(scene, fb.vp);
      fbp.update();
      repaint();
   }


   /**
      Create an instance of this class which has
      the affect of creating the GUI application.
   */
   public static void main(String[] args)
   {
      print_help_message();

      // Define initial dimensions for a FrameBuffer.
      int width  = 1024;
      int height = 1024;
      // Create an InteractiveFrame containing a FrameBuffer
      // with the given dimensions. NOTE: We need to call the
      // AntiAliasDemo constructor in the Java GUI Event
      // Dispatch Thread, otherwise we get a race condition
      // between the constructor (running in the main() thread)
      // and the very first ComponentEvent (running in the EDT).
      javax.swing.SwingUtilities.invokeLater(
         new Runnable() // an anonymous inner class constructor
         {
            public void run() // implement the Runnable interface
            {
               // call the constructor that builds the gui
               new AntiAliasDemo("Renderer 6", width, height, 10);
            }
         }
      );
   }//main()


   private static void print_help_message()
   {
      System.out.println("Use the 'a' key to toggle antialiasing on and off.");
      System.out.println("Use the 'g' key to toggle gamma correction on and off.");
      System.out.println("Use the 'c' key to change the random solid model color.");
      System.out.println("Use the 'C' key to randomly change model's colors.");
    //System.out.println("Use the 'e' key to change the random vertex colors.");
      System.out.println("Use the 'e' key to change the random solid edge colors.");
      System.out.println("Use the 'E' key to change the random edge colors.");
      System.out.println("Use the r/R keys to change the speed of rotation.");
      System.out.println("Use the f/F keys to slow down or speed up the frame rate.");
      System.out.println("Use the 'h' key to redisplay this help message.");
   }
}
