/*

*/

import renderer.scene.*;
import renderer.models.*;
import renderer.pipeline.*;
import renderer.framebuffer.*;
import renderer.gui.*;

import java.awt.Color;
import java.awt.event.*;
import java.io.File;
import java.util.List;
import java.util.ArrayList;

/**
   A rotation in three-dimensional space is specified by giving
   an angle {@code theta} and a unit vector {@code (u1, u2, u3)}
   which determines the rotation's axis.
<p>
   An important theorem in mathematics says that every rotation in
   three-dimensional space can also be specified by giving an angle
   {@code a1} of rotation around the x-axis, an angle {@code a2} of
   rotation around the y-axis, and an angle {@code a3} of rotation
   around the z-axis. That is, give angle {@code theta} and a unit
   vector {@code (u1, u2, u3)}, there are three angles {@code a1},
   {@code a2} , and {@code a3} so that
   <pre>{@code
      R(theta, u1, u2, u3) = R_z(a3) * R_y(a2) * R_x(a1)
   }</pre>
<p>
   Because of this theorem, many graphics systems do all of their
   rotations as a product of three rotations like this
   <pre>{@code
      R_z(a3) * R_y(a2) * R_x(a1)
   }</pre>
   Unfortunately, this scheme is prone to a problem called "gimbal lock".
<p>
   Suppose that we agree to always implement a rotation as three
   rotations in this specific order.
   <pre>{@code
      R_z(a3) * R_y(a2) * R_x(a1)
   }</pre>
   If the {@code y} rotation is 90 or 270, then the {@code x}
   and {@code z} rotations will be "locked" together. That is,
   the x-rotation and z-rotation parameters together only
   determine a rotation of one dimension instead of two.
<p>
   Here is a calculation that demonstrates this fact when the
   y-rotation is 90 degrees. Let {@code a1} and {@code a3} be
   arbitrary angles for the x-rotation and the z-rotation.
   <pre>{@code
   R_z(a3) * R_y(90) * R_x(a1)

   [cos(a3)  -sin(a3)  0]   [ cos(90)  0  sin(90)]   [1     0         0   ]
 = [sin(a3)   cos(a3)  0] * [    0     1     0   ] * [0  cos(a1)  -sin(a1)]
   [   0         0     1]   [-sin(90)  0  cos(90)]   [0  sin(a1)   cos(a1)]

   [cos(a3)  -sin(a3)  0]   [ 0  0  1]   [1     0         0   ]
 = [sin(a3)   cos(a3)  0] * [ 0  1  0] * [0  cos(a1)  -sin(a1)]
   [   0         0     1]   [-1  0  0]   [0  sin(a1)   cos(a1)]

   [cos(a3)  -sin(a3)  0]   [ 0  sin(a1)   cos(a1)]
 = [sin(a3)   cos(a3)  0] * [ 0  cos(a1)  -sin(a1)]
   [   0         0     1]   [-1    0          0   ]

   [ 0  sin(a1)*cos(a3)-sin(a3)*cos(a1)   cos(a1)*cos(a3)+sin(a1)*sin(a3)]
 = [ 0  cos(a1)*cos(a3)+sin(a1)*sin(a3)  -sin(a1)*cos(a3)+cos(a1)*sin(a3)]
   [-1                 0                                 0               ]

   [ 0  sin(a1 - a3)   cos(a1 - a3)]
 = [ 0  cos(a1 - a3)  -sin(a1 - a3)]
   [-1      0              0       ]
   }</pre>
<p>
   The final transformation has only one parameter, {@code a1-a3},
   so the final rotation is really a one-dimensional rotation, not
   a two-dimensional rotation.
<p>
   You can verify these calculations using this program.
   Use this program's key-commands to set the y-rotation
   to 90 degrees and then change the x and z rotations.
   You can then see how changing the x and z rotations
   affect the transformation matrix.
<p>
   It is important to realize that gimbal lock is caused by
   implementing rotations as three rotations around the x,
   y, and z axes. If rotations are implemented some other way
   (using an angle and an axis vector, or using quaternions)
   then there is no gimbal lock.
<p>
   We can create a "gimbal lock" around any axis by putting that
   axis in the middle of the three rotation transformations.
<p>
   See <a href="https://en.wikipedia.org/wiki/Gimbal_lock" target="_top">
                https://en.wikipedia.org/wiki/Gimbal_lock</a>
*/
@SuppressWarnings("serial")
public class GimbalLock extends InteractiveFrame
{
   private int xRotation = 0;
   private int yRotation = 0;
   private int zRotation = 0;

   private Scene scene;
   private List<Model> modelArray = new ArrayList<>();
   private int currentModel = 0;

   /**
      This constructor instantiates the Scene object
      and initializes it with appropriate geometry.
   */
   public GimbalLock(String title, int fbWidth, int fbHeight)
   {
      super(title, fbWidth, fbHeight);

      // Create the Scene object that we shall render.
      scene = new Scene();

      // Create a Position in the Scene.
      Position position = new Position();

      // Push the position away from where the camera is.
      position.matrix = Matrix.translate(0, 0, -2);

      // Add the Position to the Scene.
      scene.addPosition(position);

      // Create several Model objects.
      modelArray.add(new ObjSimpleModel(new File("../assets/cessna.obj")));
      modelArray.add(new ObjSimpleModel(new File("../assets/galleon.obj")));
      modelArray.add(new ObjSimpleModel(new File("../assets/cow.obj")));
      modelArray.add(new ObjSimpleModel(new File("../assets/horse.obj")));

      // Give the models random colors.
      for (Model m : modelArray)
      {
         ModelShading.setRandomColor(m);
      }

      // Add a Model to the Position.
      position.model = modelArray.get(0);
   }


   // Implement part of the KeyListener interface.
   @Override public void keyTyped(KeyEvent e)
   {
      //System.out.println( e );

      char c = e.getKeyChar();
      if ('h' == c)
      {
         print_help_message();
         return;
      }
      else if ('a' == c)
      {
         RasterizeAntialias.doAntialiasing = ! RasterizeAntialias.doAntialiasing;
         System.out.print("Anti-aliasing is turned ");
         System.out.println(RasterizeAntialias.doAntialiasing ? "On" : "Off");
      }
      else if ('g' == c)
      {
         RasterizeAntialias.doGamma = ! RasterizeAntialias.doGamma;
         System.out.print("Gamma correction is turned ");
         System.out.println(RasterizeAntialias.doGamma ? "On" : "Off");
      }
      else if ('p' == c)
      {
         scene.camera.perspective = ! scene.camera.perspective;
         String p = scene.camera.perspective ? "perspective" : "orthographic";
         System.out.println("Using " + p + " projection");
      }
      else if ('/' == c)
      {
         currentModel = (currentModel + 1) % modelArray.size();
         scene.positionList.get(0).model = modelArray.get(currentModel);
      }
      else if ('?' == c)
      {
         currentModel = (currentModel - 1);
         if (currentModel < 0) currentModel = modelArray.size() - 1;
         scene.positionList.get(0).model = modelArray.get(currentModel);
      }
      else if ('c' == c)
      {
         // Change the solid random color of the model.
         ModelShading.setRandomColor(scene.positionList.get(0).model);
      }
      else if ('C' == c)
      {
         // Change each color in the model to a random color.
         ModelShading.setRandomColors(scene.positionList.get(0).model);
      }
      else if ('e' == c && e.isAltDown())
      {
         // Change the random color of each vertex of the model.
         ModelShading.setRandomVertexColors(scene.positionList.get(0).model);
      }
      else if ('e' == c)
      {
         // Change the solid random color of each edge of the model.
         ModelShading.setRandomLineSegmentColors(scene.positionList.get(0).model);
      }
      else if ('E' == c)
      {
         // Change the random color of each end of each edge of the model.
         ModelShading.setRainbowLineSegmentColors(scene.positionList.get(0).model);
      }
      else if ('r' == c)
      {
         xRotation = 0;
         yRotation = 0;
         zRotation = 0;
      }
      else if ('x' == c)
      {
         xRotation -= 5.0;
         if (-360 == xRotation) xRotation = 0;
      }
      else if ('X' == c)
      {
         xRotation += 5.0;
         if (360 == xRotation) xRotation = 0;
      }
      else if ('y' == c)
      {
         yRotation -= 5.0;
         if (-360 == yRotation) yRotation = 0;
      }
      else if ('Y' == c)
      {
         yRotation += 5.0;
         if (360 == yRotation) yRotation = 0;
      }
      else if ('z' == c)
      {
         zRotation -= 5.0;
         if (-360 == zRotation) zRotation = 0;
      }
      else if ('Z' == c)
      {
         zRotation += 5.0;
         if (360 == zRotation) zRotation = 0;
      }

      // Set the model-to-view transformation matrix.
      // The order of the transformations is very important!
      Position model_p = scene.positionList.get(0);
      model_p.matrix2Identity();
      // Push the model away from where the camera is.
      model_p.matrix.mult( Matrix.translate(0, 0, -2) );
      // Move the model relative to its new position.
      model_p.matrix.mult( Matrix.rotateZ(zRotation) );
      model_p.matrix.mult( Matrix.rotateY(yRotation) );//the "gimbal" can lock on the middle axis
      model_p.matrix.mult( Matrix.rotateX(xRotation) );

      if ('x'==c||'X'==c||'y'==c||'Y'==c||'z'==c||'Z'==c||'r'==c)
      {
         System.out.println("xRot = " + xRotation
                        + ", yRot = " + yRotation
                        + ", zRot = " + zRotation);
         System.out.println( model_p.matrix );
      }

      // Render again.
      FrameBuffer fb = this.fbp.getFrameBuffer();
      fb.clearFB(Color.black);
      Pipeline.render(scene, fb.vp);
      fbp.update();
      repaint();
   }


   // Implement part of the ComponentListener interface.
   @Override public void componentResized(ComponentEvent e)
   {
      //System.out.println( e );

      // Get the new size of the FrameBufferPanel.
      int w = this.fbp.getWidth();
      int h = this.fbp.getHeight();

      // Create a new FrameBuffer that fits the new window size.
      FrameBuffer fb = new FrameBuffer(w, h);
      this.fbp.setFrameBuffer(fb);
      fb.clearFB(Color.black);
      Pipeline.render(scene, fb.vp);
      fbp.update();
      repaint();
   }


   /**
      Create an instance of this class which has
      the affect of creating the GUI application.
   */
   public static void main(String[] args)
   {
      print_help_message();

      // Define initial dimensions for a FrameBuffer.
      int width  = 1024;
      int height = 1024;
      // Create an InteractiveFrame containing a FrameBuffer
      // with the given dimensions. NOTE: We need to call the
      // GimbalLock constructor in the Java GUI Event
      // Dispatch Thread, otherwise we get a race condition
      // between the constructor (running in the main() thread)
      // and the very first ComponentEvent (running in the EDT).
      javax.swing.SwingUtilities.invokeLater(
         new Runnable() // an anonymous inner class constructor
         {
            public void run() // implement the Runnable interface
            {
               // call the constructor that builds the gui
               new GimbalLock("Gimbal Lock", width, height);
            }
         }
      );
   }//main()


   private static void print_help_message()
   {
      System.out.println("Use the '/' key to cycle through the models.");
      System.out.println("Use the 'p' key to toggle between parallel and orthographic projection.");
      System.out.println("Use the x/X, y/Y, z/Z, keys to rotate the models around the x, y, z axes.");
      System.out.println("Use the 'c' key to change the random solid model color.");
      System.out.println("Use the 'C' key to randomly change model's colors.");
    //System.out.println("Use the 'e' key to change the random vertex colors.");
      System.out.println("Use the 'e' key to change the random solid edge colors.");
      System.out.println("Use the 'E' key to change the random edge colors.");
      System.out.println("Use the 'a' key to toggle antialiasing on and off.");
      System.out.println("Use the 'g' key to toggle gamma correction on and off.");
      System.out.println("Use the 'r' key to reset the transformation matrix to the identity.");
      System.out.println("Use the 'h' key to redisplay this help message.");
      System.out.println();
      System.out.println("Set the y-rotation to 90 degrees to create gimbal lock.");
      System.out.println();
   }
}
