/*

*/

import renderer.scene.*;
import renderer.models.*;
import renderer.pipeline.*;
import renderer.framebuffer.*;

import java.awt.Color;

/**
   Rotate the unit square around its center, with its center at the origin.
   <pre>{@code
       y                           y                           y

       |                           |                           |
       |                           |                           |
     1 +---------+               1 +                         1 +
       |         |                 |                           |
       |         |                 |                           |
       |    +    |            +----|----+                     /|\
       |         |            |    |    |                   /  |  \
       |         |            |    |    |                 /    |    \
  -----+---------+--> x    ---+----+----+----+--> x    --<-----+----->----+-> x
       |         1            |    |    |    1            \    |    /     1
       |                      |    |    |                   \  |  /
       |                      +----+----+                     \|/
       |                           |                           |
   }</pre>
<p>
   See the animation <a href="../Example_04.gif">Example_04.gif</a>.
<p>
   <a href="../Example_04.gif"><img src="../Example_04.gif" alt="Example_04.gif"></a>
<p>
   For the i'th frame of the animation, the square's model matrix is
   <pre>{@code
      M = R(6*i) * T(-0.5, -0.5, 0)
        = R(6) * [R(6*(i-1)) * T(-0.5, -0.5, 0)].
   }</pre>
   Notice how new rotations are being accumulated on the left end of the formula.
<p>
   In this example, we need to take the square model and translate it down 1/2 unit
   and to the left 1/2 unit and then rotate counter-clockwise.
   <pre>{@code
       y                           y                           y

       |                           |                           |
       |                           |                           |
     1 +---------+               1 +                         1 +
       |         |                 |                           |
       |         |                 |                           |
       |    +    |            +----|----+                     /|\
       |         |            |    |    |                   /  |  \
       |         |            |    |    |                 /    |    \
  -----+---------+--> x    ---+----+----+----+--> x    --<-----+----->----+-> x
       |         1            |    |    |    1            \    |    /     1
       |                      |    |    |                   \  |  /
       |                      +----+----+                     \|/
       |                           |                           |
   }</pre>
<p>
   Let us analyze how we use matrix transformations to implement
   this animation.
<p>
   Let {@code M} denote the model matrix for the square. Initially,
   {@code M} is the identity matrix.
   <pre>{@code
               M = [ 1  0  0  0 ]
                   [ 0  1  0  0 ]
                   [ 0  0  1  0 ]
                   [ 0  0  0  1 ]
   }</pre>
<p>
   Let {@code R(a)} denote the 4x4 homogeneous matrix that rotates the
   xy-plane around the z-axis by angle {@code a} degrees counter-clockwise.
   <pre>{@code
            R(a) = [ cos(a * (pi/180))   -sin(a * (pi/180))   0   0 ]
                   [ sin(a * (pi/180))    cos(a * (pi/180))   0   0 ]
                   [       0                    0             1   0 ]
                   [       0                    0             0   1 ]
   }</pre>
<p>
   Let {@code T(a,b,c)} denote the 4x4 homogeneous matrix that translate by
   the amount {@code a} in the x-direction, the amount {@code b} in the
   y-direction, and the amount {@code c} in the z-direction.
   <pre>{@code
        T(a,b,c) = [ 1  0  0  a ]
                   [ 0  1  0  b ]
                   [ 0  0  1  c ]
                   [ 0  0  0  1 ]
   }</pre>
<p>
   To create the first frame of the animation, we want the square's model
   matrix to translate the square down half a unit and to the left half a
   unit.
   <pre>{@code
       M = T(-0.5, -0.5, 0)
   }</pre>
<p>
   To create the second frame of the animation, we want the square's model
   matrix to translate the square down and to the left and then rotate by
   6 degrees.
   <pre>{@code
       M = R(6) * T(-0.5, -0.5, 0)
   }</pre>
   Notice how the rotation matrix is on the left of the translation matrix.
   We want each vertex of the square model to be first translated and then
   rotated. That is, if {@code v} is a vertex from the square model we want
   <pre>{@code
      M * v = R(6) * T(-0.5, -0.5, 0) * v
   }</pre>
<p>
   For the third frame of the animation, we want the square's model matrix
   to first translate and then rotate by 12 degree.
   <pre>{@code
       M = R(6) * R(6) * T(-0.5, -0.5, 0)
         = R(12) * T(-0.5, -0.5, 0)
   }</pre>
   Notice again that the rotation matrix is multiplied on the left.
<p>
   For the {@code i}'th frame of the animation we want the square's model
   matrix to first translate and then rotate by {@code 6*1} degrees.
   <pre>{@code
       M = R(6*i) * T(-0.5, -0.5, 0)
   }</pre>
   Notice how the rotations are accumulating on the left of the square's
   model matrix. This might make you think that the loop for this animation
   would look like this.
   <pre>{@code
      for (int i = 0; i <= 60; i++)
      {
         square_p.matrix = Matrix.identity();
         square_p.mult( Matrix.translate(-0.5, -0.5, 0) );
         square_p.mult( Matrix.rotateZ(6*i) );

         fb.clearFB(Color.black);
         Pipeline.render(scene, fb.vp);
         fb.dumpFB2File(String.format("PPM_Example_Frame%03d.ppm", i));
      }
   }</pre>
   This code seems to say, "set the model matrix to the identity matrix,
   then translate, then rotate". But let us read the code very carefully.
   The first line in the body of the loop.
   <pre>{@code
      square_p.matrix = Matrix.identity();
   }</pre>
   means
   <pre>{@code
      M = I
   }</pre>
   Recall that the second line of code
   <pre>{@code
      square_p.mult( Matrix.translate(-0.5, -0.5, 0) );
   }</pre>
   means the following formula (with the translation matrix multiplied on
   the right).
   <pre>{@code
      M = M * T(-0.5, -0.5, 0)
        = I * T(-0.5, -0.5, 0)
        = T(-0.5, -0.5, 0)
   }</pre>
   The next line of code,
   <pre>{@code
      square_p.mult( Matrix.rotateZ(6*i) );
   }</pre>
   then means (with the rotation matrix multiplied on the right)
   <pre>{@code
      M = M * R(6*i)
        = T(-0.5, -0.5, 0) * R(6*i)
   }</pre>
   But this is the opposite of what we want. Above we said that the model matrix
   should translate then rotate, which has the formula
   <pre>{@code
      M = R(6*i) * T(-0.5, -0.5, 0)
   }</pre>
   Here is what the loop for creating the animation should look like.
   <pre>{@code
      for (int i = 0; i <= 60; i++)
      {
         square_p.matrix = Matrix.identity();
         square_p.mult( Matrix.rotateZ(6*i) );
         square_p.mult( Matrix.translate(-0.5, -0.5, 0) );

         fb.clearFB(Color.black);
         Pipeline.render(scene, fb.vp);
         fb.dumpFB2File(String.format("PPM_Example_Frame%03d.ppm", i));
      }
   }</pre>
   If you read the code "backwards", from the bottom up, it says "translate
   and then rotate", which is what we want (when we read the formula from
   right-to-left). But if you read the code forwards, from the top down, it
   seems to say "rotate then translate". There is a way to make sense of
   both of these ways of reading the code and the second way of reading the
   code, from the top down, is actually the preferred way to read it once
   you understand how to interpret it.
<p>
   We know that the code
   <pre>{@code
      square_p.matrix = Matrix.identity();
      square_p.mult( Matrix.rotateZ(6*i) );
      square_p.mult( Matrix.translate(-0.5, -0.5, 0) );
   }</pre>
   creates the following model matrix because of the rule about multiplying
   matrices on the right.
   <pre>{@code
      M = I * R(6*i) * T(-0.5, -0.5, 0)
   }</pre>
   Notice how, if you read the code from the bottom up, you are reading
   the formula from right-to-left. But if you read the code from the top
   down, you are reading the formula from left-to-right. Let {@code v} be
   a vertex from the square model. When we apply the model matrix to the
   vertex, we get this formula,
   <pre>{@code
      M * v = I * R(6*i) * T(-0.5, -0.5, 0) * v
            = R(6*i) * ( T(-0.5, -0.5, 0) * v )
   }</pre>
   which says that the vertex is first multiplied by the translation, then
   the result of the translation is multiplied by the rotation. In other
   words, "translate then rotate". So the "bottom up" (or right-to-left)
   way of reading the code tells us what the code does to each vertex.
<p>
   Here is how to read the code from the top down (or, how to read the
   formula from left-to-right). When read from the top down, we interpret
   each transformation as transforming the coordinate system. So
   <pre>{@code
      square_p.mult( Matrix.rotateZ(6*i) );
   }</pre>
   means rotate the coordinate system. And
   <pre>{@code
      square_p.mult( Matrix.translate(-0.5, -0.5, 0) );
   }</pre>
   means, translate the <em>rotated</em> coordinate system. And then the
   vertices of the square get plotted in the translated version of the
   rotated coordinate system.
<p>
   The next example program visualizes this transformation of the coordinate system.
<p>
   Here is the scene tree that this program builds for the i'th frame
   of the animation. There is only one Matrix object in the scene. All
   of the transformation information needs to be in that single matrix.
   Notice how the additional transformation needed by each new frame, a
   rotation by 6 degrees, is accumulated on the left of the position's
   matrix. Since multiplication "on the left" is not the way matrix
   multilication is implemented in most graphics libraries, this
   example is a bit more complicated than the previous example.
   <pre>{@code
             Scene
            /     \
           /       \
      Camera       List<Position>
                         |
                         |
                      Position
                      /       \
                     /         \
                Matrix          Square
          R(6i)*T(-1/2,-1/2)
   }</pre>
<p>
   (The code below does show a way to "force" the matrix
   multiplcation to be on the left of the position's matrix.)
*/
public class Example_04a
{
   public static void main(String[] args)
   {
      // Create the Scene object that we shall render.
      Scene scene = new Scene();

      // Create a FrameBuffer to render our scene into.
      int width  = 512;
      int height = 512;
      FrameBuffer fb = new FrameBuffer(width, height);

      // Create a set of x and y axes.
      Model axes = new Axes2D(-2, +2, -2, +2, 8, 8);
      // Color them red.
      ModelShading.setColor(axes, Color.red);
      // Create a Position for the axes.
      Position axes_p = new Position(axes);
      // Add the axes to the Scene.
      scene.addPosition(axes_p);
      // Push the axes away from where the camera is.
      axes_p.matrix = Matrix.translate(0, 0, -2);

      // Create a Model of a square.
      Model square = new Model();
      square.addVertex(new Vertex(0, 0, 0),
                       new Vertex(1, 0, 0),
                       new Vertex(1, 1, 0),
                       new Vertex(0, 1, 0));
      square.addColor(Color.magenta,
                      Color.green,
                      Color.magenta,
                      Color.green);
      square.addLineSegment(new LineSegment(0, 1, 0, 1),
                            new LineSegment(1, 2, 1, 2),
                            new LineSegment(2, 3, 2, 3),
                            new LineSegment(3, 0, 3, 0));

      // Create a Position for the square.
      Position square_p = new Position(square);
      // Add the square to the Scene.
      scene.addPosition(square_p);

      // Rotate the square around its center,
      // with its center at the origin.
      for (int i = 0; i < 60; i++)
      {
         // Push the square away from where the camera is.
         square_p.matrix = Matrix.translate(0, 0, -2);

         // Rotate the square.
         square_p.matrix.mult( Matrix.rotateZ(6*i) ); // notice the difference here
         // Translate the square's center to the origin, (0,0).
         square_p.matrix.mult( Matrix.translate(-0.5, -0.5, 0) );

         // Render again.
         fb.clearFB(Color.darkGray);
         Pipeline.render(scene, fb.vp);
         fb.dumpFB2File(String.format("PPM_Example_04a_Frame%03d.ppm", i));
      }



      // Here is a way to "force" the rotations to accumulate
      // on the left of the Position's model matrix.
      // Push the square away from where the camera is.
      square_p.matrix = Matrix.translate(0, 0, -2);
      // Translate the square's center to the origin, (0,0).
      square_p.matrix.mult( Matrix.translate(-0.5, -0.5, 0) );

      // Rotate the square around its center,
      // with its center at the origin.
      for (int i = 0; i < 60; i++)
      {
         // Render again.
         fb.clearFB(Color.darkGray);
         Pipeline.render(scene, fb.vp);
         fb.dumpFB2File(String.format("PPM_Example_04av2_Frame%03d.ppm", i));

         // Accumulate rotations on the left of the model matrix.
         square_p.matrix = Matrix.rotateZ(6).times(square_p.matrix);
      }
   }
}
